import ctypes
kernel32 = ctypes.windll.kernel32
user32 = ctypes.windll.user32
from pathlib import Path
import logging
import io
import sys
import os
import traceback
import threading
import weakref


def HideCMD():
    whnd = kernel32.GetConsoleWindow()
    if whnd != 0:
        user32.ShowWindow(whnd, 0)


def ShowCMD():
    whnd = kernel32.GetConsoleWindow()
    if whnd != 0:
        user32.ShowWindow(whnd, 4)
        # win32con.SW_SHOWNOACTIVATE


def NoBlockCMD():
    '''取消或禁用控制台/终端的快速编辑模式'''
    kernel32.SetConsoleMode(kernel32.GetStdHandle(-10), (0x4 | 0x80 | 0x20 | 0x2 | 0x10 | 0x1 | 0x00 | 0x100))


def InitConsole():
    NoBlockCMD()
    HideCMD()
    ctypes.windll.shcore.SetProcessDpiAwareness(1)
    InitWin32api()


def InitWin32api():
    import pywintypes
    import win32api
    import win32con


class StreamToLogger(io.TextIOBase):
    '''缓存接受到的日志文本，并在接受到换行符或者缓冲命令时输出。
    不具有实时性'''

    def __init__(self, logger, level, ):
        self.logger = logger
        self.level = level
        self.msg = ""

    def write(self, s: str):
        self.msg += s
        lines = self.msg.splitlines()
        if len(lines) > 1:
            self.msg = lines.pop()
            for line in lines:
                try:
                    self.logger.log(self.level, line.rstrip(' \r\n'))
                except:
                    pass
        return len(s)

    def flush(self):
        if len(self.msg) > 0:
            self.write(self.msg + '\n')
            self.msg = ""


class Route_log(io.TextIOBase):
    def __init__(self):
        self.handler = weakref.WeakSet()

    def write(self, s: str) -> int:
        for func in self.handler:
            try:
                func.write(s)
            except:
                pass
        return len(s)

    def flush(self) -> None:
        for func in self.handler:
            try:
                func.flush()
            except:
                pass
        return

    def __del__(self):
        self.handler.clear()
        del self.handler


Route_stdout = Route_log()
Route_stdout.handler.add(sys.__stdout__)
Route_stderr = Route_log()
Route_stderr.handler.add(sys.__stderr__)


try:
    OutputFolder = Path(__file__).parent.parent / 'output'
    OutputFolder.mkdir(exist_ok=True)
except:
    OutputFolder = Path(os.environ['TEMP']) / 'output'
    OutputFolder.mkdir(exist_ok=True)

GLB_INFO = None
GLB_ERROR = None


def Init_logging():
    OutputFile = OutputFolder / 'LOG1.log'
    selfdef_fmt = '%(asctime)s - %(funcName)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(selfdef_fmt)

    logging.basicConfig(
        filename=OutputFile,
        filemode='a+',
        format=selfdef_fmt,
        encoding='utf-8',
        level=logging.DEBUG
    )
    log = logging.getLogger()

    sys.stdout = Route_stdout
    sys.stderr = Route_stderr

    global GLB_INFO, GLB_ERROR

    GLB_INFO = StreamToLogger(log, logging.INFO)
    GLB_ERROR = StreamToLogger(log, logging.ERROR)
    Route_stdout.handler.add(GLB_INFO)
    Route_stderr.handler.add(GLB_ERROR)

    pass


if __name__ == "__main__":

    print('输出至标准输出1')
    print('输出至标准错误2', file=sys.stderr)
    Init_logging()

    print('输出至标准输出3')
    print('输出至标准错误4', file=sys.stderr)

    logging.debug('This message should go to the log file')
    logging.info('So should this')
    logging.warning('And this, too')
    logging.error('And non-ASCII stuff, too, like Øresund and Malmö')

    try:
        1 / 0
    except Exception:
        logging.exception("Exception occurred")

    print('asdasda')
    print('MYERROR', file=sys.stderr)


pass
