from typing import Dict, List, Any, TypeAlias, TypeVar, Literal, Callable
import os
import sys
import clr
import io
import pythoncom
import threading
import traceback
import logging
from tkinter import messagebox

_cs_path = [
    # r'E:\Seafile\CS6_Proj\WLXWindow\bin\Debug'
]

_cs_name = [
    'System',
    # 'System.Runtime',
    'System.Collections',
    'WLXWindow',
    'PresentationFramework',
    'Microsoft.WindowsAPICodePack.Shell'
]

# 添加Dll文件的搜索路径 加载对应名称的dll文件
list(map(lambda x: sys.path.append(x), _cs_path))
list(map(lambda x: clr.AddReference(x), _cs_name))


import System    # pyright: ignore
import System.Reflection as Reflection     # pyright: ignore
from System.Collections import Generic    # pyright: ignore
from System.Windows import MessageBox, MessageBoxButton, MessageBoxImage, MessageBoxResult  # pyright: ignore
from Microsoft.Win32 import OpenFileDialog   # pyright: ignore
from Microsoft.WindowsAPICodePack.Dialogs import CommonOpenFileDialog, CommonFileDialogResult    # pyright: ignore
import WLXWindow  # pyright: ignore


class CSList(Generic.List):
    def __new__(cls, clstype, _iterable):
        self = cls[clstype]()
        for x in _iterable:
            self.Add(x)
        return self


class ObservableCollection_T(System.Collections.ObjectModel.ObservableCollection):
    def __new__(cls, clstype, _iterable):
        self = cls[clstype]()
        for x in _iterable:
            self.Add(x)
        return self


class ErrorHint:

    def __enter__(self):
        return self

    def __exit__(self, exec_type, exec_value, exec_tb):
        if exec_type is not None:
            logging.exception("Exception occurred")
            print(exec_type, exec_value, exec_tb)
            if exec_type is ModuleNotFoundError or exec_type is ImportError:
                messagebox.showinfo(title='警告', message="导入模块出错,请重新执行\n模块安装.bat")
                return False
            else:
                errortext = io.StringIO()
                traceback.print_exception(*sys.exc_info(), file=errortext)
                # messagebox.showinfo(title='警告', message=f"运行中出现异常错误,生成终止\n{errortext.getvalue()}")
                messageBoxText = f"运行中出现异常错误,生成终止\n{errortext.getvalue()}"
                result = MessageBox.Show(messageBoxText, "警告", MessageBoxButton.OK, MessageBoxImage.Warning)
        else:
            pass
        return True


class Win_log(io.TextIOBase):
    def __init__(self, TextBoxObj) -> None:
        self.tx = TextBoxObj
        r1, w1 = os.pipe()
        self.read_side = os.fdopen(r1, 'r')
        self.write_side = os.fdopen(w1, 'w')
        self.td = threading.Thread(target=self.loop, args=(), kwargs=None, daemon=True)
        self.td.start()

    def write(self, s: str) -> int:
        ret = self.write_side.write(s)
        self.write_side.flush()
        return ret

    def update(self, s):
        self.tx.AppendText(s)
        self.tx.ScrollToEnd()

    def loop(self):
        msg = ""
        c = ""
        updateAction = System.Action[str](self.update)
        while True:
            try:
                c = self.read_side.read(1)
                if c == '':
                    pass
                    break
                else:
                    msg += c
                if c in '\r\n':
                    self.tx.Dispatcher.BeginInvoke(updateAction, msg)
                    msg = ""
                elif len(msg) >= 512:
                    self.tx.Dispatcher.BeginInvoke(updateAction, msg)
                    msg = msg[512:]
            except:
                traceback.print_exception(*sys.exc_info())
                pass


def SetSTAThread():
    pythoncom.CoInitializeEx(pythoncom.COINIT_APARTMENTTHREADED)


class SingleActionError(Exception):
    pass


class SingleAction(threading.Thread):
    _running = dict()

    def __init__(self, target: Callable[..., object], target_name: str, args=(), kwargs: Dict[str, Any] | None = None) -> None:
        super().__init__(group=None, target=target, name=None, args=args, kwargs=kwargs, daemon=True)
        if isinstance(target_name, str):
            self.target_name = target_name
        else:
            self.target_name = target.__name__

    def run(self):
        state = SingleAction._running.get(self.target_name, None)
        if state == 1:
            del self._target, self._args, self._kwargs
            return
        else:
            SingleAction._running[self.target_name] = 1

        SetSTAThread()
        try:
            if self._target is not None:
                self._target(*self._args, **self._kwargs)
        except:
            logging.exception("Exception occurred")
        finally:
            del self._target, self._args, self._kwargs
            SingleAction._running.pop(self.target_name, 0)
