

import ConsoleApi
from CsharpApi import *

import ctypes
from ctypes import wintypes

import re
import csv
import struct
import time
import copy
import traceback
import logging
from pathlib import Path
import collections

import tkinter
from tkinter import filedialog
from typing import List, Dict
from datetime import datetime

import win32api
import win32con
import win32gui
import win32process
import commctrl


from commctrl import (LVM_GETITEMCOUNT, LVM_GETITEMTEXT, LVM_GETNEXTITEM, LVNI_SELECTED)
from win32gui import (FindWindow, FindWindowEx, SendMessage, GetCursorPos)
from win32process import (GetWindowThreadProcessId, ReadProcessMemory, VirtualAllocEx, VirtualFreeEx, WriteProcessMemory)

import pywintypes
from pywintypes import HANDLEType

from ctypes import wintypes, WINFUNCTYPE
from win32con import WH_KEYBOARD, WH_KEYBOARD_LL, WH_MOUSE, WH_MOUSE_LL, HC_ACTION
from win32con import WM_KEYDOWN, WM_KEYUP
from win32con import KF_EXTENDED, KF_ALTDOWN, KF_REPEAT, KF_UP

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32

CallNextHookEx = user32.CallNextHookEx
UnhookWindowsHookEx = user32.UnhookWindowsHookEx
SetWindowsHookEx = user32.SetWindowsHookExW


LRESULT = wintypes.LPARAM

Hookproc_ptr = WINFUNCTYPE(LRESULT, wintypes.INT, wintypes.WPARAM, wintypes.LPARAM)
SetWindowsHookEx.argtypes = (wintypes.INT, Hookproc_ptr, wintypes.HINSTANCE, wintypes.DWORD)
SetWindowsHookEx.restype = wintypes.HHOOK

CallNextHookEx.argtypes = (wintypes.HHOOK, wintypes.INT, wintypes.WPARAM, wintypes.LPARAM)
CallNextHookEx.restype = LRESULT

SendMessageA = user32.SendMessageA
SendMessageA.restype = LRESULT

user32.WindowFromPoint.argtypes = [wintypes.POINT]
user32.WindowFromPoint.restype = wintypes.HWND

user32.GetComboBoxInfo.argtypes = [wintypes.HWND, wintypes.HWND]
user32.GetComboBoxInfo.restype = wintypes.BOOL


class Structure_walk:
    '''遍历结构体'''

    def __init__(self, struct_obj):
        self.struct_obj = struct_obj
        self.stack = []
        self.dump = {}

    def walk_struct(self, struct_obj):
        for field_name, field_type, *field_other in struct_obj._fields_:
            field_value = getattr(struct_obj, field_name)
            self.stack.append(field_name)
            if isinstance(field_value, (ctypes.Structure, ctypes.Union)):
                self.walk_struct(field_value)
            else:
                self.dump['.'.join(self.stack)] = field_value
            self.stack.pop()

    def __str__(self):
        self.stack.clear()
        self.dump.clear()
        self.walk_struct(self.struct_obj)
        return str(self.dump)


class Structure_tool:
    def __str__(self) -> str:
        return str(Structure_walk(self))
    __repr__ = __str__


class myc_uint(ctypes._SimpleCData):
    _type_ = "I"


class C32:
    UINT = ctypes.c_uint32
    INT = ctypes.c_int32
    LPARAM = ctypes.c_int32
    WPARAM = ctypes.c_uint32
    HWND = ctypes.c_uint32
    LONG = ctypes.c_int32
    ULONG = ctypes.c_uint32
    INT_PTR = ctypes.c_int32
    LPSTR = ctypes.c_uint32
    PUINT = ctypes.c_uint32
    PINT = ctypes.c_uint32

    class POINT(wintypes.POINT, Structure_tool):
        pass


class C32_LVITEMA(ctypes.Structure, Structure_tool):
    # region member_define
    mask: int
    iItem: int
    iSubItem: int
    state: int
    stateMask: int
    pszText: int
    cchTextMax: int
    iImage: int
    lParam: int
    iIndent: int
    iGroupId: int
    cColumns: int
    puColumns: int
    piColFmt: int
    iGroup: int
    # endregion

    _fields_ = [
        ("mask", C32.UINT),
        ("iItem", C32.INT),
        ("iSubItem", C32.INT),
        ("state", C32.UINT),
        ("stateMask", C32.UINT),
        ("pszText", C32.LPSTR),
        ("cchTextMax", C32.INT),
        ("iImage", C32.INT),
        ("lParam", C32.LPARAM),
        ("iIndent", C32.INT),
        ("iGroupId", C32.INT),
        ("cColumns", C32.UINT),
        ("puColumns", C32.PUINT),
        ("piColFmt", C32.PINT),
        ("iGroup", C32.INT),
    ]


class MOUSEHOOKSTRUCT(ctypes.Structure, Structure_tool):
    # region 注释
    pt: wintypes.POINT
    hwnd: int
    wHitTestCode: int
    dwExtraInfo: int
    # endregion
    _fields_ = [("pt", wintypes.POINT),
                ("hwnd", wintypes.HWND),
                ("wHitTestCode", wintypes.UINT),
                ("dwExtraInfo", wintypes.WPARAM),
                ]


class MSLLHOOKSTRUCT(ctypes.Structure, Structure_tool):
    # region 注释
    pt: wintypes.POINT
    mouseData: int
    flags: int
    time: int
    dwExtraInfo: int
    # endregion

    _fields_ = [("pt", wintypes.POINT),
                ("mouseData", wintypes.DWORD),
                ("flags", wintypes.DWORD),
                ("time", wintypes.DWORD),
                ("dwExtraInfo", wintypes.WPARAM),
                ]


class KBDLLHOOKSTRUCT(ctypes.Structure, Structure_tool):
    # region 注释
    vkCode: wintypes.POINT
    scanCode: int
    flags: int
    time: int
    dwExtraInfo: int
    # endregion

    _fields_ = [("vkCode", wintypes.DWORD),
                ("scanCode", wintypes.DWORD),
                ("flags", wintypes.DWORD),
                ("time", wintypes.DWORD),
                ("dwExtraInfo", wintypes.WPARAM),
                ]


class C32_COMBOBOXEXITEMA(ctypes.Structure, Structure_tool):
    # region member_define
    mask: int
    iItem: int
    pszText: int
    cchTextMax: int
    iImage: int
    iSelectedImage: int
    iOverlay: int
    iIndent: int
    lParam: int
    # endregion

    _fields_ = [
        ('mask', C32.UINT),
        ('iItem', C32.INT_PTR),
        ('pszText', C32.LPSTR),
        ('cchTextMax', C32.INT),
        ('iImage', C32.INT),
        ('iSelectedImage', C32.INT),
        ('iOverlay', C32.INT),
        ('iIndent', C32.INT),
        ('lParam', C32.LPARAM)
    ]


class COMBOBOXINFO(ctypes.Structure, Structure_tool):
    # region member_define
    cbSize: int
    rcItem: wintypes.RECT
    rcButton: wintypes.RECT
    stateButton: int
    hwndCombo: int
    hwndItem: int
    hwndList: int
    # endregion
    _fields_ = [
        ('cbSize', wintypes.DWORD),
        ('rcItem', wintypes.RECT),
        ('rcButton', wintypes.RECT),
        ('stateButton', wintypes.DWORD),
        ('hwndCombo', wintypes.HWND),
        ('hwndItem', wintypes.HWND),
        ('hwndList', wintypes.HWND)
    ]


def GetComboBoxInfo(hwnd):
    temp_cbo = COMBOBOXINFO()
    temp_cbo.cbSize = ctypes.sizeof(temp_cbo)
    ret = user32.GetComboBoxInfo(hwnd, ctypes.addressof(temp_cbo))
    return temp_cbo


def MAKEWPARAM(l, h):
    return (l & 0x0000_FFFF) | ((h & 0x0000_FFFF) << 16)


class C32_RAM:
    def __init__(self, hProcess: HANDLEType, dwSize: int, flAllocationType=win32con.MEM_COMMIT, flProtect=win32con.PAGE_READWRITE):
        pass
        self.hProcess = hProcess
        self.ptr = VirtualAllocEx(hProcess, False, dwSize, flAllocationType, flProtect)
        self.size = dwSize

    @property
    def data(self) -> bytes:
        return ReadProcessMemory(self.hProcess, self.ptr, self.size)

    @data.setter
    def data(self, _value):
        if isinstance(_value, (bytes, bytearray)):
            d_size = len(_value)
            if d_size <= self.size:
                bytes_written = WriteProcessMemory(self.hProcess, self.ptr, bytes(_value))
            else:
                self.release()
                C32_RAM.__init__(self, self.hProcess, d_size)

    def release(self):
        VirtualFreeEx(self.hProcess, self.ptr, 0, win32con.MEM_RELEASE)

    def __del__(self):
        try:
            self.release()
        except:
            pass


class C32_RAM_STR(C32_RAM):

    def __init__(self, hProcess: HANDLEType, init: bytes | int | str, encoding='gbk'):
        if isinstance(init, (bytes, bytearray)):
            C32_RAM.__init__(self, hProcess, len(init))
            init_data = bytes(init)

        elif isinstance(init, int):
            C32_RAM.__init__(self, hProcess, init)
            # init_data = bytes(init)

        elif isinstance(init, str):
            init_data = init.encode(encoding) + b'\0'
            C32_RAM.__init__(self, hProcess, len(init_data))
        else:
            raise TypeError(init)

        if isinstance(init, (bytes, bytearray, str)):
            bytes_written = WriteProcessMemory(self.hProcess, self.ptr, init_data)
        self.encoding = encoding

    @property
    def text(self) -> str:
        return self.read_text_from(self.ptr)

    @text.setter
    def text(self, _value):
        if isinstance(_value, str):
            data = _value.encode(self.encoding) + b'\0'
            self.data = data
        raise ValueError(_value)

    def read_text_from(self, address):
        ret: bytes = ReadProcessMemory(self.hProcess, address, self.size)
        ret_text = ret.decode(encoding=self.encoding, errors='replace').rstrip('\0')
        return ret_text


def int_to_hex(integer):
    hex_str = hex(integer)[2:]  # Convert integer to hexadecimal string
    hex_str = hex_str.zfill(8)  # Pad the string with zeros to 8 characters
    return hex_str


def GetSpecifyItemString(MMEProcessHandle: int, hwnd: int, row: int = 0, column: int = 0, bufferSize=512) -> str:
    '''获取字符串内容\n
    HANDLE MMEProcessHandle :进程句柄, \n
    HWND hwnd :控件句柄, \n
    int row,\n
    int column, \n
    int bufferSize
    '''

    test_LVITEM = C32_LVITEMA()
    size = ctypes.sizeof(C32_LVITEMA)

    # 在表格对象中创建一个内存空间
    ram_LVITEM = C32_RAM_STR(MMEProcessHandle, init=size)
    RAM_stringBuffer = C32_RAM_STR(MMEProcessHandle, init=bufferSize, encoding='gbk')
    before_ptr = RAM_stringBuffer.ptr

    # 定义当前选中的表格元素
    test_LVITEM.cchTextMax = bufferSize
    test_LVITEM.iSubItem = column
    test_LVITEM.pszText = RAM_stringBuffer.ptr
    buffer = bytes(test_LVITEM)

    # 按照指定的初始结构初始化申请的内存空间
    ram_LVITEM.data = buffer

    # # 从指定行获取 LVITEM 对象 ,存放至申请的空间
    SendMessageA(hwnd, LVM_GETITEMTEXT, row, ram_LVITEM.ptr)

    # 从对象中获取字符串数据,
    test_LVITEM2 = C32_LVITEMA.from_buffer_copy(ram_LVITEM.data)

    if test_LVITEM2.pszText == test_LVITEM.pszText:
        ret_text = RAM_stringBuffer.text
        return ret_text
    elif test_LVITEM2.pszText > 0:
        return RAM_stringBuffer.read_text_from(test_LVITEM2.pszText)
    else:
        return ''


def SetItemState(MMEProcessHandle: int, hwnd: int, row: int = 0):
    '''将某一行设置为选中状态'''
    test_LVITEM = C32_LVITEMA()
    size = ctypes.sizeof(C32_LVITEMA)

    # 在表格对象中创建一个内存空间
    ram_LVITEM = C32_RAM_STR(MMEProcessHandle, size)

    # 定义当前选中的表格元素
    test_LVITEM.iItem = row
    test_LVITEM.state = commctrl.LVIS_SELECTED
    test_LVITEM.stateMask = commctrl.LVIS_SELECTED
    buffer = bytes(test_LVITEM)

    # 按照指定的初始结构初始化申请的内存空间
    ram_LVITEM.data = buffer

    # # 从指定行获取 LVITEM 对象 ,存放至申请的空间
    SendMessageA(hwnd, commctrl.LVM_SETITEMSTATE, row, ram_LVITEM.ptr)

    # 从对象中获取字符串数据
    return test_LVITEM


def ComboBox_GetLBText(hwndCtl: int, index: int = 0) -> str:
    text_len = SendMessageA(hwndCtl, win32con.CB_GETLBTEXTLEN, index, 0)
    buffer = ctypes.create_string_buffer(text_len + 1)
    ret = SendMessageA(hwndCtl, win32con.CB_GETLBTEXT, index, buffer)

    lret = ctypes.c_int32(ret).value

    if lret >= 0:
        ret_text = buffer.value.decode('gbk')
        return ret_text
    else:
        return ''


def ComboBox_SetText(hwndCtl: int, text: str, encoding='gbk') -> str:
    buffer = ctypes.create_string_buffer(text.encode(encoding=encoding))
    SendMessageA(hwndCtl, win32con.WM_SETTEXT, 0, buffer)
    ComboBox_FreshText(hwndCtl)


# def ComboBox_SetClipboardText(hwndCtl: int, text: str) -> str:
#     win32clipboard.OpenClipboard(0)
#     win32clipboard.EmptyClipboard()
#     win32clipboard.SetClipboardText(text)
#     try:
#         win32clipboard.CloseClipboard()
#         pass
#     except pywintypes.error:
#         logging.exception('Exception occurred')

#     SendMessageA(hwndCtl, win32con.WM_PASTE, 0, 0)
#     ComboBox_FreshText(hwndCtl)


def ComboBox_SendUnicode(hwndCtl: int, text: str) -> str:
    for c in text:
        SendMessage(hwndCtl, win32con.WM_IME_CHAR, ord(c), 1)
    ComboBox_FreshText(hwndCtl)


# def ComboBox_SELECTSTRING():
    # # 可以修改内容 但是在进行替换操作时还是获取 最后交互时的设置
    # # buffer = ctypes.create_string_buffer("整个工程文件".encode(encoding='gbk'))

    # SendMessageA(dlg.Cbo_窗口选择, win32con.WM_LBUTTONDOWN, 0, 0)
    # SendMessageA(dlg.Cbo_窗口选择, win32con.WM_LBUTTONUP, 0, 0)

    # buffer = ctypes.create_string_buffer("窗口 116".encode(encoding='gbk'))
    # ret = SendMessageA(dlg.Cbo_窗口选择, win32con.CB_SELECTSTRING, -1, buffer)

def ComboBox_FreshItem(hwndCtl: int):
    cb_id = win32gui.GetWindowLong(hwndCtl, win32con.GWL_ID)
    ret = SendMessage(win32gui.GetParent(hwndCtl), win32con.WM_COMMAND, MAKEWPARAM(cb_id, win32con.CBN_SELCHANGE), hwndCtl)
    return ret


def ComboBox_FreshText(hwndCtl: int):
    cb_id = win32gui.GetWindowLong(hwndCtl, win32con.GWL_ID)
    ret = SendMessage(win32gui.GetParent(hwndCtl), win32con.WM_COMMAND, MAKEWPARAM(cb_id, win32con.CBN_EDITCHANGE), hwndCtl)
    return ret


def ComboBox_FreshText(hwndCtl: int):
    cb_id = win32gui.GetWindowLong(hwndCtl, win32con.GWL_ID)
    ret = SendMessage(win32gui.GetParent(hwndCtl), win32con.WM_COMMAND, MAKEWPARAM(cb_id, win32con.CBN_EDITCHANGE), hwndCtl)
    return ret


def ComboBox_SetCurSel(hwndCtl: int, index):
    SendMessageA(hwndCtl, win32con.CB_SETCURSEL, index, 0)
    ComboBox_FreshItem(hwndCtl)


def Button_Click(hwndCtl: int):
    btn_id = win32gui.GetWindowLong(hwndCtl, win32con.GWL_ID)
    ret = SendMessage(win32gui.GetParent(hwndCtl), win32con.WM_COMMAND, MAKEWPARAM(btn_id, win32con.BN_CLICKED), hwndCtl)
    return ret


def get_combobox_text(hwnd) -> List[str]:
    '''获取组合框控件的可选项内容'''
    MMECombox = hwnd
    threadId, processId = MMEComboxPid = GetWindowThreadProcessId(MMECombox)
    MMEProcessHandle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, processId)

    # 获取组合框项目数量
    count = SendMessage(MMECombox, win32con.CB_GETCOUNT, 0, 0)
    print('当前组合框的总行元素量', count)

    cb_list = []
    # 获取组合框每一项文本
    if count > 0:
        for i in range(0, count):
            窗口名 = ComboBox_GetLBText(MMECombox, i)
            # print(窗口名)
            cb_list.append(窗口名)
    else:
        print('列表无数据')
    return cb_list


EB_fieldnames = ('位置', '名称', '地址', '寻找值', '替换值', '忽略', '指定窗口')


class INVO_IOmap(collections.UserList):
    '''UserList用户自定义列表类
    对类的 列表类型的操作:会被重定向到self.data
    对类的 属性属性的操作:和平常一样
    只需要一个self.data = []就算初始化完成
    '''

    def __init__(self, newlist=None) -> None:
        super().__init__(initlist=newlist)

    @staticmethod
    def _get_data_index(index):
        def _infunc(selfobj):
            return selfobj.data[index]
        return _infunc

    @staticmethod
    def _set_data_index(index):
        def _infunc(selfobj, _value):
            selfobj.data[index] = _value
        return _infunc

    var_name: str = property(_get_data_index(0), _set_data_index(0))
    entrance: str = property(_get_data_index(1), _set_data_index(1))
    description: str = property(_get_data_index(3), _set_data_index(3))
    iec_address: str = property(_get_data_index(4), _set_data_index(4))
    device_name: str = property(_get_data_index(5), _set_data_index(5))


class EBSearchDialog(object):
    '''尝试获取EasyBuilder的对话框'''
    MMEWindow: int

    default_csv = r'D:\eb_replace.csv'

    def __new__(cls, *args, **kwargs):
        try:
            MMEWindow = FindWindow("#32770", "寻找/替换")
            if MMEWindow == 0:
                return None
        except:
            logging.exception('未找到窗口')
        self = super().__new__(cls)
        self.MMEWindow = MMEWindow
        return self

    def __init__(self, *args, **kwargs) -> None:

        MMEWindow = self.MMEWindow

        self.Cbo_窗口选择 = FindWindowEx(MMEWindow, None, "ComboBox", None)

        #    //2、通过父窗口句柄获取 控件的句柄
        self.Tab_地址页 = FindWindowEx(MMEWindow, None, "#32770", None)
        地址页 = self.Tab_地址页

        # 第一个设备标记
        设备标记 = FindWindowEx(地址页, None, "Static", "设备 :")
        前标签 = FindWindowEx(地址页, None, "Static", "双击鼠标, 可检视元件")

        self.Cbo_寻找值_设备 = FindWindowEx(地址页, 设备标记, "ComboBox", None)
        self.Cbo_寻找值_标签 = FindWindowEx(地址页, 前标签, "ComboBox", None)

        设备标记2 = FindWindowEx(地址页, 设备标记, "Static", "设备 :")

        self.Cbo_替换值_设备 = FindWindowEx(地址页, 设备标记2, "ComboBox", None)
        self.Cbo_替换值_标签 = FindWindowEx(地址页, self.Cbo_寻找值_标签, "ComboBox", None)

        self.Chk_验证标签名称 = FindWindowEx(地址页, None, "Button", "验证标签名称")

        self.Btn_寻找 = FindWindowEx(地址页, None, "Button", "寻找")
        self.Btn_替换 = FindWindowEx(地址页, self.Btn_寻找, "Button", "替换")
        self.Btn_全部替换 = FindWindowEx(地址页, None, "Button", "全部替换")

        self.ListView_搜索结果 = FindWindowEx(地址页, None, "SysListView32", "")

        threadId, processId = GetWindowThreadProcessId(MMEWindow)
        self.threadId = threadId
        self.processId = processId

        self.csvpath = EBSearchDialog.default_csv

        self.lastsize = 0

        self.alltext: List[Dict[str, str]]

        self.no_record = False

        pass

    def get_search_result(self, save=False, auto_arrange=False) -> List[Dict[str, str]]:
        '''获取当前搜索结果中的显示数据'''
        MMEListViewProcess = None

        try:
            MMEListView = self.ListView_搜索结果
            MMEListViewProcess = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, self.processId)

            # //5、获取表格总的数量
            count = SendMessage(MMEListView, LVM_GETITEMCOUNT, 0, 0)
            print('当前列表的总行元素量', count)
            # //6、获取当前被选中的行号,未选择时为 4294967295 不是 -1,行号从0开始
            selected = SendMessage(MMEListView, LVM_GETNEXTITEM, -1, LVNI_SELECTED)
            # 将所有的字符串输出
            # [['无法发现', '"LB-?"', '', '', '', '']]

            alltext = []
            if count > 0:
                fieldnames = EB_fieldnames
                for i in range(0, count):
                    位置 = GetSpecifyItemString(MMEListViewProcess, MMEListView, i, 0)
                    名称 = GetSpecifyItemString(MMEListViewProcess, MMEListView, i, 1)
                    地址 = GetSpecifyItemString(MMEListViewProcess, MMEListView, i, 2)

                    if 位置 == '':
                        continue

                    temp_data = [位置, 名称, 地址]
                    temp_data.extend(['' for _ in range(len(fieldnames) - len(temp_data))])

                    temp_dict = dict(zip(EB_fieldnames, temp_data))
                    alltext.append(temp_dict)
            else:
                print('当前列表内容为空')

            self.alltext = alltext

            if len(alltext) > 0:
                if save:
                    if auto_arrange:
                        def sorting_function(x):
                            value: str = x['位置']
                            if value.startswith('窗口'):
                                try:
                                    win_index = int(value.replace('窗口_', ''))
                                    value = f'窗口_{win_index:>04d}'
                                except:
                                    pass
                            # return value, x['名称']
                            return value

                        alltext.sort(key=sorting_function)

                    with open(file=self.csvpath, mode='w+', encoding='gbk', errors='replace') as csvfile:
                        writer = csv.DictWriter(csvfile, fieldnames=fieldnames, lineterminator='\n')
                        # 写表头
                        writer.writeheader()
                        # 写数据
                        writer.writerows(alltext)
            else:
                print('无可用数据或显示"无法发现"')

            pass
        except:
            logging.exception('异常')
        finally:
            if MMEListViewProcess:
                win32api.CloseHandle(MMEListViewProcess)

        return alltext

    def set_item_state(self, search: str, result):
        '''获取当前搜索结果中的显示数据'''
        MMEListViewProcess = None
        try:
            MMEListView = self.ListView_搜索结果
            MMEListViewProcess = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, self.processId)
            count = SendMessage(MMEListView, LVM_GETITEMCOUNT, 0, 0)

            searched_flag = False
            for i, data in enumerate(result, start=0):
                if i > count:
                    break
                if search == data['名称']:
                    ret = SetItemState(MMEListViewProcess, MMEListView, i)
                    searched_flag = True
                    print(f'第{i}项搜索结果,被选中')
            if not searched_flag:
                print('没有匹配项')

        except:
            logging.exception('异常')
        return

    def replace_as_csv(self):
        '''根据csv文件中的设置执行替换操作'''
        # 检查是否能够输入
        if not win32gui.IsWindowVisible(self.Tab_地址页):
            print('未选择正确的地址页面')
            return
        if not win32gui.IsWindowVisible(self.Cbo_寻找值_标签) \
                or not win32gui.IsWindowVisible(self.Cbo_替换值_标签):
            print('未选择正确的PLC设备')
            return

        # 从csv文件中载入数据
        rootfile = filedialog.askopenfilename()
        if len(rootfile) > 0:
            csvpath = rootfile
        else:
            # raise Exception('文件未选择')
            return

        try:
            with open(file=csvpath, mode='r', encoding='gbk', errors='replace') as csvfile:
                reader = csv.DictReader(csvfile)
                head = reader.fieldnames
                alltext = list(reader)
        except:
            print('表格文件打开失败')
            return

        self.lastalltext = alltext

        replace_error = []
        ignore_win = '指定窗口' not in head
        try:
            # 获取可选的窗口列表
            eb_windows = get_combobox_text(self.Cbo_窗口选择)
            keyword = []
            for title in eb_windows:
                # 0~3为特殊设置,后面的才是窗口
                keyword.append(title.split(' : ')[0])

            alltext_len = len(alltext)
            for i, action in enumerate(alltext, start=0):
                print(f'处理第 {i+1}/{alltext_len}')
                if action['寻找值'] == '':
                    print(i, '未填写查找值,跳过')
                    continue

                if action['替换值'] == '':
                    print(i, action, '警告,替换值为空')
                    continue

                if len(action['忽略']) > 0:
                    print(i, '设置了忽略单元格,跳过')
                    continue

                ComboBox_SetText(self.Cbo_寻找值_标签, action['寻找值'])
                ComboBox_SetText(self.Cbo_替换值_标签, action['替换值'])

                if not ignore_win and action['指定窗口'] in ('', '0', None):
                    # 不修改窗口指定值
                    pass
                if not ignore_win and action['指定窗口'] == '1':
                    # 指定 整个工程文件
                    ComboBox_SetCurSel(self.Cbo_窗口选择, 0)

                if not ignore_win and action['指定窗口'] in ['2', '3']:
                    # 设置在指定窗口上查找数据
                    if action['位置'] in ['报警 (事件) 登录', '资料取样']:
                        ComboBox_SetCurSel(self.Cbo_窗口选择, 2)
                        pass
                    elif action['位置'].find('窗口') > -1:

                        try:
                            # 设置新的选择序号    指定要选择的字符串的从零开始的索引。 如果此参数为 -1，则删除列表中的任何当前选定内容并清除编辑控件。
                            eb_index = keyword.index(action['位置'].replace('_', ' '))
                            ComboBox_SetCurSel(self.Cbo_窗口选择, eb_index)
                            pass
                        except ValueError:
                            print(action['位置'], '无法设置')
                            continue
                        except:
                            logging.exception('异常')
                            continue

                if action['指定窗口'] == '3':
                    single_flag = 1
                else:
                    single_flag = 0

                # 执行替换操作
                # 单击寻找按钮
                Button_Click(self.Btn_寻找)
                time.sleep(0.2)

                # 检查是否有搜索到的数据
                result_1 = self.get_search_result()

                # 检测能否全部替换
                ret = win32gui.IsWindowEnabled(self.Btn_全部替换)

                # 以单选模式替换时执行以下操作
                if single_flag == 1:
                    pass
                    # 定位并选中名称匹配的行
                    self.set_item_state(action['名称'], result_1)

                    # 单击替换对选中的内容进行替换
                    Button_Click(self.Btn_替换)

                else:
                    if ret == 1:
                        # 按下全部替换
                        Button_Click(self.Btn_全部替换)

                        # 检查替换后的数据
                        result_2 = self.get_search_result()

                        if len(result_1) != len(result_2):
                            print(f'第{i}项,数据长度不一致?')
                        else:
                            for ri in range(len(result_1)):
                                d1 = result_1[ri]
                                d2 = result_2[ri]

                                if d1['地址'] == d2['地址']:
                                    print(f'第{i}项', d1, '项', '替换失败,替换前后的地址相同')
                                    print(d1)
                                    replace_error.append(d1)

                    else:
                        print(f'第{i}项无替换选择,余下{len(alltext)-i-1}')

                time.sleep(0.2)
                # 更新记录

                if not self.no_record:
                    action['忽略'] = '1'

                x, y = GetCursorPos()
                if ((0 <= x < 10) and (0 <= y < 10)):
                    print('特殊结束')
                    return

            print(f'处理正常完成')
        except:
            logging.exception('异常')
        finally:
            print(f'开始回写结果')
            try:
                with open(file=csvpath, mode='w+', encoding='gbk', errors='replace') as csvfile:
                    writer = csv.DictWriter(csvfile, fieldnames=head, lineterminator='\n')
                    # 写表头
                    writer.writeheader()
                    # 写数据
                    writer.writerows(alltext)

                if len(replace_error) > 0:
                    error_csv = Path(csvpath).with_name(f'eb_error_record{datetime.now():%m%d%H%M}.csv')
                    with open(file=error_csv, mode='w+', encoding='gbk', errors='replace') as csvfile:
                        writer = csv.DictWriter(csvfile, fieldnames=head, lineterminator='\n')
                        # 写表头
                        writer.writeheader()
                        # 写数据
                        writer.writerows(replace_error)
            except:
                print('保存处理数据失败')

        pass


def change_number() -> None:
    '''尝试将原有的5位数字转为从0开始的单独数字'''

    # 从csv文件中载入数据
    rootfile = filedialog.askopenfilename()
    if len(rootfile) > 0:
        csvpath = rootfile
    else:
        # raise Exception('文件未选择')
        return

    try:
        with open(file=csvpath, mode='r', encoding='gbk', errors='replace') as csvfile:
            reader = csv.DictReader(csvfile)
            head = reader.fieldnames
            alltext = list(reader)
    except:
        print('表格文件打开失败')
        return

    def repl_func(objMatch: re.Match) -> str:
        # temp_text = objMatch.re.sub(string=objMatch.group(0), repl=_repl),
        aa = objMatch.groupdict()['aa']

        num = objMatch.groupdict()['num'].strip('[]')
        num = int(num[3:]) - 1
        logging.info(f"{objMatch.groupdict()['num']} -> [{num}]")
        return f'{aa}[{num}]'

    for i, action in enumerate(alltext, start=0):

        sText = action['地址']
        objPattern = re.compile(r'''(?P<aa>\w+)(?P<num>\[\d{5}\])''', re.M | re.I)
        sText = objPattern.sub(repl=repl_func, string=sText)

        objPattern = re.compile(r'''(?P<full>VAR://)''', re.M)
        sText = objPattern.sub(repl=r'Application.GVL.', string=sText)

        action['寻找值'] = sText

    # 回写数据
    csvpath = Path(csvpath).with_stem('eb_default_change')
    with open(file=csvpath, mode='w+', encoding='gbk', errors='replace') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=head, lineterminator='\n')
        # 写表头
        writer.writeheader()
        # 写数据
        writer.writerows(alltext)


def change_number_v2() -> Path:
    '''尝试将原有的5位数字转为从0开始的单独数字'''

    # 从csv文件中载入数据
    rootfile = filedialog.askopenfilename()
    if len(rootfile) > 0:
        csvpath = rootfile
    else:
        # raise Exception('文件未选择')
        return

    try:
        with open(file=csvpath, mode='r', encoding='gbk', errors='replace') as csvfile:
            reader = csv.DictReader(csvfile)
            head = reader.fieldnames
            alltext = list(reader)
    except:
        print('表格文件打开失败')
        return

    i = 0
    action = None
    head_record = set()

    def repl_func(objMatch: re.Match) -> str:
        finded_text = objMatch.group(0)

        aa = objMatch.groupdict()['aa']
        # aa = 'Application.GVL.'
        bb = objMatch.groupdict()['bb']
        num = objMatch.groupdict()['num'].strip('[]')
        num = int(num[3:]) - 1
        logging.info(f"{objMatch.groupdict()['num']} -> [{num}]")

        ret_text = f'{aa}{bb}[{num}]'

        if finded_text not in head_record:
            head_record.add(finded_text)
            action['寻找值'] = finded_text
            action['替换值'] = ret_text.replace('VAR://', 'Application.GVL.')
        return

    for i, action in enumerate(alltext, start=0):
        sText = action['地址']

        # 搜索带有 [10101]形式 5位数字的地址文本
        objPattern = re.compile(r'''(?P<aa>VAR://)(?P<bb>.{1,150})(?P<num>\[\d{5}\])''', re.M | re.I)
        sText = objPattern.sub(repl=repl_func, string=sText)

    # 回写数据
    csvpath = Path(csvpath).with_stem('eb_default_change')
    with open(file=csvpath, mode='w+', encoding='gbk', errors='replace') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=head, lineterminator='\n')
        # 写表头
        writer.writeheader()
        # 写数据
        writer.writerows(alltext)
    return csvpath


def resize_window(handle, width, height):
    win32gui.MoveWindow(handle, 0, 0, width, height, True)


class Win(WLXWindow.EBProToolDialog):

    def __init__(self, *args) -> None:
        '''  '''
        WLXWindow.EBProToolDialog.__init__(self, *args)
        self.Txt_Log = self.FindName("Txt_Log")
        self.Txt_LogOut = Win_log(self.Txt_Log)

        self.Tab_TopMenu = self.FindName("Tab_TopMenu")

        self.Btn_SelectSourceFile = self.FindName("Btn_SelectSourceFile")
        self.Cbo_Source = self.FindName("Cbo_Source")

        self.Btn_Generate = self.FindName("Btn_Generate")
        self.Cbk_LaterOpen = self.FindName("Cbk_LaterOpen")
        self.Cbk_AutoArrange = self.FindName("Cbk_AutoArrange")
        self.Cbk_SetShareText = self.FindName("Cbk_SetShareText")

        self.Btn_SetChange = self.FindName("Btn_SetChange")

        self.Btn_FBExport = self.FindName("Btn_FBExport")
        self.Cbk_RecordHistory = self.FindName("Cbk_RecordHistory")

        self.Cbo_Source.Text = EBSearchDialog.default_csv

        self.Btn_Generate.Click += self.save_search_result
        self.Btn_SelectSourceFile.Click += self.SelectFile
        self.Btn_SetChange.Click += self.change_dt

        self.Btn_FBExport.Click += self.start_replace
        self.Loaded += self.callback1
        self.Closed += self.unhook

        # self.cmButton = self.FindResource("cmButton")
        self.Mitem_1.Click += self.save_search_result
        self.Mitem_2.Click += self.start_replace

        self.ff_smc2 = ('csv文件', '*.csv')
        self.ff_default = self.ff_smc2
        self.dlg: EBSearchDialog = None
        self.replacing = False
        self.Hook_HANDLE = 0

        pass

    def callback1(self, sender, e):
        if self.Hook_HANDLE == 0:

            whwnd = System.Windows.Interop.WindowInteropHelper(self).Handle.ToInt64()
            threadId, processId = win32process.GetWindowThreadProcessId(whwnd)

            self.KeyboardProc = Hookproc_ptr(self.PyKeyboardProc)
            self.MouseProc = Hookproc_ptr(self.PyMouseProc)

            mode = 4
            if mode == 1:
                self.Hook_MODE = WH_KEYBOARD
                self.Hook_HANDLE = SetWindowsHookEx(self.Hook_MODE, self.KeyboardProc, 0, threadId)
            elif mode == 2:
                self.Hook_MODE = WH_KEYBOARD_LL
                self.Hook_HANDLE = SetWindowsHookEx(self.Hook_MODE, self.KeyboardProc, 0, 0)
            elif mode == 3:
                self.Hook_MODE = WH_MOUSE
                self.Hook_HANDLE = SetWindowsHookEx(self.Hook_MODE, self.MouseProc, 0, threadId)
            elif mode == 4:
                self.Hook_MODE = WH_MOUSE_LL
                self.Hook_HANDLE = SetWindowsHookEx(self.Hook_MODE, self.MouseProc, 0, 0)

            print(f'\n挂钩ID: {self.Hook_HANDLE} {self.Hook_HANDLE:016X}')

            if isinstance(self.Hook_HANDLE, int) and self.Hook_HANDLE != 0:
                print('挂钩成功')
                logging.info(f'\n挂钩ID: {self.Hook_HANDLE} {self.Hook_HANDLE:016X}')
            else:
                print('挂钩失败')
                self.Hook_HANDLE = 0

    def PyMouseProc(self, nCode: int, wParam: int, lParam: int):

        if nCode == HC_ACTION:
            if self.Hook_MODE == WH_MOUSE:
                MS_Data = MOUSEHOOKSTRUCT.from_address(lParam)
                # logging.info((nCode, wParam, MS_Data))
                # print((nCode, wParam, MS_Data))
            elif self.Hook_MODE == WH_MOUSE_LL:
                # MS_Data = MSLLHOOKSTRUCT.from_address(lParam)
                # logging.info((nCode, wParam, MS_Data))
                # print(hwndPointNow)
                # print((nCode, wParam, MS_Data))
                if wParam == win32con.WM_RBUTTONUP:
                    MS_Data = MSLLHOOKSTRUCT.from_address(lParam)
                    hwndPointNow = user32.WindowFromPoint(MS_Data.pt)
                    hwndRoot = win32gui.GetAncestor(hwndPointNow, win32con.GA_ROOT)
                    WindowTitle = win32gui.GetWindowText(hwndRoot)

                    if WindowTitle == '寻找/替换':
                        self.ShowContextMenuFromNonUIThread()
                    # print(hwndPointNow, hwndRoot, ClassName)

        return CallNextHookEx(self.Hook_HANDLE, nCode, wParam, lParam)

    def PyKeyboardProc(self, nCode: int, wParam: int, lParam: int):

        if nCode == HC_ACTION:
            if self.Hook_MODE == WH_KEYBOARD:
                keyFlags = (lParam & 0xFFFF_0000) >> 16
                KB_Data = {
                    'vkCode': wParam,
                    'lParam': f"{lParam:08X}",
                    'repeatCount': lParam & 0x0000FFFF,
                    # 0-15	重复计数。 该值是用户按住键后重复击键的次数。
                    'ScanCode': (lParam & 0x00FF0000) >> 16,
                    # 16-23 扫描代码。 此值取决于 OEM。
                    # 'keyFlags': (lParam & 0xFFFF_0000) >> 16,
                    'isExtendedKey': (keyFlags & KF_EXTENDED) == KF_EXTENDED,
                    # 24	指示键是扩展键，例如功能键还是数字键盘上的键。 如果键是扩展键，则值为 1;否则为 0。
                    'isAltDown': (keyFlags & KF_ALTDOWN) == KF_ALTDOWN,
                    # 29	上下文代码。 如果 Alt 键关闭，则值为 1;否则为 0。
                    'isRepeat': (keyFlags & KF_REPEAT) == KF_REPEAT,
                    # 30	上一个键状态。 如果在发送消息之前键关闭，则值为 1;如果键已打开，则为 0。
                    'isKeyUp': (keyFlags & KF_UP) == KF_UP,
                    # 31	转换状态。 如果按下键，则值为 0;如果释放键，则值为 1。
                }
                # logging.info((nCode, wParam, KB_Data))
                # print((nCode, wParam, KB_Data))
            elif self.Hook_MODE == WH_KEYBOARD_LL:
                KB_Data = KBDLLHOOKSTRUCT.from_address(lParam)
                # logging.info((nCode, wParam, KB_Data))
                # print((nCode, wParam, KB_Data))
        return CallNextHookEx(self.Hook_HANDLE, nCode, wParam, lParam)

    def unhook(self, sender, e):
        '''函数功能'''
        if self.Hook_HANDLE > 0:
            UnhookWindowsHookEx(self.Hook_HANDLE)

    def SelectFile(self, sender, e):
        dlg = OpenFileDialog()
        dlg.DefaultExt = self.ff_default[1]
        dlg.Filter = '|'.join(self.ff_default)
        dlg.CheckFileExists = True
        result = dlg.ShowDialog()
        if result:
            # self.xlsxmap = dlg.FileName
            self.Cbo_Source.Text = dlg.FileName
            return
        else:
            e.Handled = True
            return

    def save_search_result(self, sender, e):
        '''函数功能'''
        # self.Tab_TopMenu.SelectedIndex = 1

        dlg = EBSearchDialog()
        if dlg is None:
            return
        dlg.csvpath = self.Cbo_Source.Text
        dlg.get_search_result(save=True, auto_arrange=self.Cbk_AutoArrange.IsChecked)
        pass
        if self.Cbk_LaterOpen.IsChecked == True:
            os.startfile(self.Cbo_Source.Text)

    def start_replace(self, sender, e):
        # self.Tab_TopMenu.SelectedIndex = 1
        self.no_record = self.Cbk_RecordHistory.IsChecked

        cfg = {
            'no_record': self.no_record
        }

        def lstart_replace(no_record):
            dlg = EBSearchDialog()
            dlg.no_record = no_record
            dlg.replace_as_csv()

        act = SingleAction(lstart_replace, 'lstart_replace', args=(cfg['no_record'],))
        act.start()
        return

    def change_dt(self, sender, e):
        try:
            file = change_number_v2()
            if self.Cbk_LaterOpen.IsChecked == True:
                os.startfile(file)
            # dlg = EBSearchDialog()
            # if dlg is None:
            #     return
            # dlg.get_item_state()
            # pass

        except:
            logging.exception('Exception occurred')


def main():
    SetSTAThread()
    try:
        app = Win()
    except:
        logging.exception("Exception occurred")
        return
    # 将输出信息重定向至界面中的日志,在Txt_LogOut被销毁时自动移除
    Txt_LogOut = ConsoleApi.Route_stdout.handler.add(app.Txt_LogOut)
    try:
        # 启动窗口
        app.ShowDialog()
    except:
        logging.exception('Exception occurred')


def test_unit():
    dlg = EBSearchDialog()
    ComboBox_FreshItem(dlg.Cbo_窗口选择)
    Button_Click(dlg.Btn_寻找)
    pass


if __name__ == "__main__":
    ConsoleApi.InitConsole()
    ConsoleApi.Init_logging()
    main()
    # test_unit()
